(* Title: navigation on the calctree
   Author: Walther Neuper 2017
   (c) due to copyright terms
*)
signature CALC_TREE_NAVIGATION =
sig
  val last_onlev : CTbasic.ctree -> Pos.pos -> bool
  val exist_lev_on' : CTbasic.ctree -> Pos.pos' -> bool                (* for interface.sml *)
  val is_curr_endof_calc : CTbasic.ctree -> Pos.pos' -> bool           (* for interface.sml *)

  val par_pblobj : CTbasic.ctree -> Pos.pos -> Pos.pos  (* duplicate in ctree-basic.sml *)

  val lev_on' : CTbasic.ctree -> Pos.pos' -> Pos.pos'              (* for interface.sml *)
  val move_dn : Pos.pos -> CTbasic.ctree -> Pos.pos' -> Pos.pos'
  val move_up : Pos.pos -> CTbasic.ctree -> Pos.pos' -> Pos.pos'(* or interface.sml *)
  val movelevel_dn : Pos.pos -> CTbasic.ctree -> Pos.pos' -> Pos.pos'(*nterface.sml *)
  val movelevel_up : Pos.pos -> CTbasic.ctree -> Pos.pos' -> Pos.pos'(*nterface.sml *)
  val movecalchd_up : CTbasic.ctree -> Pos.pos' -> Pos.pos'        (* for interface.sml *)
end

(**)
structure CTnavi(**): CALC_TREE_NAVIGATION(**) =
struct
(**)
open CTbasic
open Pos

fun last_onlev pt pos = not (existpt (lev_on pos) pt);

(* find the position of the next parent which is a PblObj in ctree *)
fun par_pblobj _ [] = []
  | par_pblobj pt p =
    let
      fun par _ [] = []
        | par pt p =
          if is_pblobj (get_obj I pt p) 
          then p
          else par pt (lev_up p)
    in par pt (lev_up p) end; 

(* determine the next pos' on the same level *)
fun lev_on' _ ([], Pbl) = ([], Res)
  | lev_on' pt (p, Res) =
    if get_obj g_res pt p = get_obj g_form pt (lev_on p)(*TransitiveB*)
    then if existpt' (lev_on p, Res) pt
      then (lev_on p, Res)
      else raise ERROR ("lev_on': (p, Res) -> (p, Res) not existent, p = " ^ ints2str' (lev_on p))
    else (lev_on p, Frm)
  | lev_on' pt (p, _) =
    if existpt' (p, Res) pt
    then (p, Res)
    else raise ERROR ("lev_on': (p, Frm) -> (p, Res) not existent, p = " ^ ints2str' p);

fun exist_lev_on' pt p =
  case \<^try>\<open> lev_on' pt p \<close> of
    SOME _ => true
  | NONE => false;

(* is the pos' at the last element of a calulation _AND_ can be continued *)
fun is_curr_endof_calc _  ([],Res) = false
  | is_curr_endof_calc pt (pos as (p, _)) =
    not (exist_lev_on' pt pos) andalso get_obj g_ostate pt (lev_up p) = Incomplete;

(* move one step down into existing nodes of ctree; skip Res = Frm.nxt;
   move_dn at the end of the calc-tree raises PTREE *)
fun move_dn _ (Nd (_, ns)) ([],p_) =                                            (* root problem *)
    (case p_ of 
	     Res => raise PTREE "move_dn: end of calculation"
	   | _ =>
	     if null ns                                                     (* go down from Pbl + Met *)
	     then raise PTREE "move_dn: solve problem not started"
	     else ([1], Frm))
  | move_dn P  (Nd (_, ns)) (p :: (ps as (_ :: _)), p_) =              (* iterate to end of pos *)
    if p > length ns
    then raise PTREE "move_dn: pos not existent 2"
    else move_dn (P @ [p]) (nth p ns) (ps, p_)
  | move_dn P (Nd (c, ns)) ([p], p_) =                            (* act on last element of pos *)
    if p > length ns
    then raise PTREE "move_dn: pos not existent 3"
    else
      (case p_ of 
	      Res => 
	      if p = length ns                               (* last Res on this level: go a level up *)
	      then if g_ostate c = Complete
	        then (P, Res)
	        else raise PTREE (ints2str' P ^ " not complete 1")
	     else                        (* go to the next Nd on this level, or down into the next Nd *)
		     if is_pblnd (nth (p + 1) ns) then (P@[p + 1], Pbl)
		     else  if g_res' (nth p ns) = g_form' (nth (p + 1) ns)
		       then if (null o children o (nth (p + 1))) ns
			       then                                                   (* take the Res if Complete *) 
			         if g_ostate' (nth (p + 1) ns) = Complete 
			         then (P@[p + 1], Res)
			         else raise PTREE (ints2str' (P@[p + 1]) ^ " not complete 2")
			       else (P@[p + 1, 1], Frm)                           (* go down into the next PrfObj *)
		       else (P@[p + 1], Frm)                           (* take Frm: exists if the Nd exists *)
	   | Frm => (*go down or to the Res of this Nd*)
	     if (null o children o (nth p)) ns
	     then if g_ostate' (nth p ns) = Complete then (P @ [p], Res)
		     else raise PTREE (ints2str' (P @ [p])^" not complete 3")
	     else (P @ [p, 1], Frm)
	   | _ =>                                                                    (* is Pbl or Met *)
	     if (null o children o (nth p)) ns
	     then raise PTREE "move_dn:solve subproblem not startd"
	     else (P @ [p, 1], 
		   if (is_pblnd o hd o children o (nth p)) ns
		   then Pbl else Frm))
  | move_dn _ _ _ = raise ERROR "";

(* go one level down into ctree *)
fun movelevel_dn [] (Nd (c, ns)) ([], _) =                                     (*  root problem *)
    if is_pblobj c 
    then if null ns 
	    then raise PTREE "solve problem not started"
	    else ([1], if (is_pblnd o hd) ns then Pbl else Frm)
    else raise PTREE "pos not existent 1"
  | movelevel_dn P (Nd (_, ns)) (p :: (ps as (_ :: _)), p_) =     (* iterate towards end of pos *)
    if p > length ns
    then raise PTREE "pos not existent 2"
    else movelevel_dn (P@[p]) (nth p ns) (ps, p_)
  | movelevel_dn P (Nd (c, ns)) ([p], p_) =                         (*act on last element of pos*)
    if p > length ns
    then raise PTREE "pos not existent 3"
    else
      (case p_ of Res => 
	      if p = length ns 
	      then raise PTREE "no children"
	      else 
		      if g_branch c = TransitiveB
		      then if (null o children o (nth (p+1))) ns
			      then raise PTREE "no children"
			      else (P @ [p+1, 1], if (is_pblnd o hd o children o (nth (p+1))) ns then Pbl else Frm)
		      else if (null o children o (nth p)) ns
		        then raise PTREE "no children"
		        else (P @ [p, 1], if (is_pblnd o hd o children o (nth p)) ns then Pbl else Frm)
	    | _ =>
	      if (null o children o (nth p)) ns 
		    then raise PTREE "no children"
		    else (P @ [p, 1],                                                            (* go down *)
			    if (is_pblnd o hd o children o (nth p)) ns then Pbl else Frm))
  | movelevel_dn _ _ _ = raise ERROR "";

(* go to the previous position in ctree; regard TransitiveB *)
fun move_up _ (Nd (c, ns)) ([], p_) = (*root problem*)
    if is_pblobj c 
    then case p_ of Res => if null ns then ([], Pbl) (*Res -> Pbl (not Met)!*)
			   else ([length ns], Res)
		  | _  => raise PTREE "begin of calculation"
    else raise PTREE "pos not existent"
  | move_up P  (Nd (_, ns)) (p :: (ps as (_ :: _)), p_) =              (* iterate to end of pos *)
    if p > length ns
    then raise PTREE "pos not existent"
    else move_up (P@[p]) (nth p ns) (ps,p_)
  | move_up P (Nd (c, ns)) ([p], p_) =                            (* act on last element of pos *)
    if p > length ns
    then raise PTREE "pos not existent"
    else if is_pblnd (nth p ns)
      then
        case p_ of Res => 
		      let val nc = (length o children o (nth p)) ns
		      in
		        if nc = 0
		        then (P @ [p], Pbl)                                        (* Res -> Pbl (not Met) *)
		       else (P @ [p, nc], Res) end                                              (* go down *)
		    | _ => if p = 1 then (P, Pbl) else (P @ [p - 1], Res) 
      else
        (case p_ of Frm =>
          if p <> 1
          then (P, Frm) 
          else if is_pblobj c then (P, Pbl) else (P, Frm)
		    | Res => 
		        let val nc = (length o children o (nth p)) ns
		        in
		          if nc = 0                                                      (* cannot go down *)
		          then if g_branch c = TransitiveB andalso p <> 1
		            then (P @ [p - 1], Res)
		            else (P @ [p], Frm)
		          else (P @ [p, nc], Res) end                                           (* go down *)
		    | _ => raise ERROR "move_up, NOT is_pblnd: uncovered case pos_") 
  | move_up _ _ _ = raise ERROR "move_up: uncovered fun def."

(* go one level up in ctree; sets the position on Frm *)
fun movelevel_up _ _ (([], _)) = raise PTREE "pos not existent"               (* root problem *)
  | movelevel_up P  (Nd (_, ns)) (p :: (ps as (_ :: _)), p_) =  (* iterate towards end of pos *)
    if p > length ns
    then raise PTREE "pos not existent"
    else movelevel_up (P @ [p]) (nth p ns) (ps, p_)
  | movelevel_up P (Nd (c, ns)) ([p], _) =                      (* act on last element of pos *)
    if p > length ns then raise PTREE "pos not existent"
    else if is_pblobj c then (P, Pbl) else (P, Frm)
  | movelevel_up _ _ _ = raise ERROR "movelevel_up: uncovered fun def."

(* go to the next calc-head up in the calc-tree *)
fun movecalchd_up pt (p, Res) = (par_pblobj pt p, Pbl)
  | movecalchd_up pt (p, _) =
    if is_pblobj (get_obj I pt p) 
    then (p, Pbl)
    else (par_pblobj pt p, Pbl);

end