(* Title:  MathEngBasic/calc-tree-elem.sml
   Author: Walther Neuper 191026
   (c) due to copyright terms

Definitions for Tactic.T and Ctree.
*)
signature MODEL_DEFINITION =
sig

  type form_T
  type form_model
  type form_refs
  val form_empty : form_T

  type variant
  type variants
  type m_field = Model_Pattern.m_field
  type descriptor = Model_Pattern.descriptor
  type values

  type o_model
  type o_model_single
  val o_model_to_string : Proof.context -> o_model -> string
  val o_model_single_to_string : Proof.context -> o_model_single -> string
  val o_model_empty : o_model_single

  type i_model
  type i_model_TEST
  type i_model_single
  type i_model_single_TEST
  datatype i_model_feedback = 
      Cor of (term * (term list)) * (term * (term list))
    | Syn of TermC.as_string 
    | Typ of TermC.as_string 
    | Inc of (term * (term list))	* (term * (term list))
    | Sup of (term * (term list)) 
    | Mis of (term * term) 
    | Par of TermC.as_string
  datatype i_model_feedback_TEST = 
      Cor_TEST of (descriptor * (term list)) * (term * (term list))
    | Syn_TEST of TermC.as_string 
    | Inc_TEST of (descriptor * (term list)) * (term * (term list))
    | Sup_TEST of (descriptor * (term list)) 

  val i_model_empty : i_model_single
end

(**)                   
structure Model_Def(**): MODEL_DEFINITION(**) =        
struct
(**)

type m_field = Model_Pattern.m_field;
type descriptor = Model_Pattern.descriptor;

(** definitions for Formalise.T **)

type form_model = TermC.as_string list;
type form_refs = References_Def.T;
type form_T = form_model * form_refs;
val form_empty = ([], References_Def.empty);

type variant = int;
type variants = variant list; (* variants given by Formalise.T *)
type values = term list;

(** definitions for O_Model.T **)

type o_model_single =
  (int *       (* sequence of input, 0 = untouched, drop with PIDE *)
	variants *   (* pointers to variants given in Formalise.T        *)
	m_field *    (* #Given | #Find | #Relate + TODO: #Where          *)
	descriptor * (* see Input_Descript.thy                           *)
	values       (* isalist2list t | [t]: the values of in item      *)
	);
type o_model = o_model_single list;  
val o_model_empty = (0, [], "", TermC.empty, [TermC.empty]) : o_model_single;

fun o_model_single_to_string ctxt (i, vs, fi, t, ts) = 
  "(" ^ string_of_int i ^ ", " ^ (strs2str o (map string_of_int)) vs ^ ", " ^ fi ^ ", " ^
  UnparseC.term ctxt t ^ ", " ^ (strs2str o (map (UnparseC.term ctxt))) ts ^ ")";
fun o_model_to_string ctxt o_model =
  (strs2str' o (map (linefeed o (o_model_single_to_string ctxt)))) o_model;


(** definitions for I_Model.T **)

datatype i_model_feedback = 
  Cor of (descriptor * (* correct input: descriptor, see Input_Descript.thy *)
     (values)) *              (* set of values for list-types: for decdiing
                                 Inc the values packed out from the list;
                                 They are packed back to an isalist in env  *)
   (term * (term list))       (* will be eliminated in i_model_feedback_TEST*)
| Syn of TermC.as_string      (* PIDE: contains "(__=__)" etc               *)
| Typ of TermC.as_string      (* type error: drop with PIDE ?               *)
| Inc of (descriptor * values) * (term * (term list))  (* see Cor           *)
| Sup of (descriptor * values)(* superfluous; input not found in O_Model.T  *)
| Mis of (descriptor * term)  (* missing; for feedback                      *)
| Par of TermC.as_string;     (* internal state from fun parsitm            *)

(*
  this datatype assumes that all items of the Model_Pattern are output
  by the initial template, where the items are displayed with hints for the input format.
*)
datatype i_model_feedback_TEST = 
  Cor_TEST of (descriptor * (* correct input: descriptor, see Input_Descript.thy *)
     (values)) *              (* set of values for list-types: for decdiing
                                 Inc the values packed out from the list;
                                 They are packed back to an isalist for envs     *)
   (term * (term list))       (* will be eliminated in i_model_feedback_TEST     *)
| Syn_TEST of TermC.as_string (* syntax error kept for P_Model.switch_pbl_met    *)
| Inc_TEST of (descriptor * values) * (term * (term list))  (* see Cor           *)
| Sup_TEST of (descriptor * values)(* superfluous; input not found in O_Model.T  *)

type i_model_single = 
  int *             (* sequence of input, 0 = untouched, drop with PIDE ?  *)
  variants *        (* pointers to variants given in Formalise.model       *)
  bool *            (* input on this item is not/complete                  *)
  m_field *         (* #Given | #Find | #Relate                            *)
  i_model_feedback; (* variables and values, see above                     *)
type i_model_single_TEST = 
  int *                     (* sequence of input, 0 = untouched, drop with PIDE ? *)
  variants *                (* pointers to variants given in Formalise.model      *)
  bool *                    (* this item is not/complete, completed by O_Model.T  *)
  m_field *                 (* #Given | #Find | #Relate                           *)
  (i_model_feedback_TEST *  (* variables and values, see above                    *)
    Position.T);            (* for pushing feedback back to PIDE                  *)

type i_model = i_model_single list;
type i_model_TEST = i_model_single_TEST list;

val i_model_empty = (0, [], false, "i_model_empty", Syn "i_model_empty");

(**)end(**)
